#include <assert.h>
#include <math.h>
#include <stdint.h>

// Packed structure for manipulating encoded numbers

typedef struct _packedFloat {
  unsigned int significand:23;
  unsigned int exponent:8;
  unsigned int sign:1;
} packedFloat;

typedef union _mix {
  float f;
  packedFloat b;
} mix;


// Unpacked structure for actual use

typedef struct _bitpattern {
  unsigned int nan:1;
  unsigned int inf:1;
  unsigned int zero:1;
  unsigned int subnormal:1;

  unsigned int sign:1;
  int exponent:10;              // Bias removed
  unsigned int significand:24;  // Leading 1 added
} unpackedFloat;


// Rounding modes

#define RNE 0
#define RNA 1
#define RTP 2
#define RTN 3
#define RTZ 4


// Note that C's == does not give bitwise equality
int compareFloat (float f, float g) {
#ifdef CUT
  // For signed zeros
  mix mf;
  mix mg;

  mf.f = f;
  mg.f = g;

  return ((mf.b == mg.b) || (isnan(f) && isnan(g)));
#else
  return ((f == g) || (isnan(f) && isnan(g)));
#endif
}

// Functions for working with unpacked floats
void initUnpackedFloat (unpackedFloat *uf) {
  uf->nan = 1;
  uf->inf = 1;
  uf->zero = 1;
  uf->subnormal = 1;

  uf->sign = 1;
  uf->exponent = -512;
  uf->significand = 0xFFFFFF;
  return;
}

void makeZero (unpackedFloat *uf) {
  uf->nan = 0;
  uf->inf = 0;
  uf->zero = 1;
  uf->subnormal = 0;

  uf->exponent = 0;
  uf->significand = 0;
  return;
}

void makeInf (unpackedFloat *uf) {
  uf->nan = 0;
  uf->inf = 1;
  uf->zero = 0;
  uf->subnormal = 0;

  uf->exponent = 0xFF;
  uf->significand = 0;
  return;
}

void makeMax (unpackedFloat *uf) {
  uf->nan = 0;
  uf->inf = 0;
  uf->zero = 0;
  uf->subnormal = 0;

  uf->exponent = 127;
  uf->significand = 0xFFFFFF;
  return;
}

void makeNaN (unpackedFloat *uf) {
  uf->nan = 1;
  uf->inf = 0;
  uf->zero = 0;
  uf->subnormal = 0;

  uf->exponent = 0xFF;
  uf->significand = 0;
}


#define BIAS 127

// Invariants for the use of the unpacked float
void check (unpackedFloat uf) {
  assert(((unsigned int)uf.nan + 
          (unsigned int)uf.inf +
          (unsigned int)uf.zero +
          (unsigned int)uf.subnormal) <= 1);

  if (uf.nan == 1) {
    assert(uf.exponent == 0xFF);
    assert(uf.significand == 0x0);

  } else if (uf.inf == 1) {
    assert(uf.exponent == 0xFF);
    assert(uf.significand == 0x0);

  } else if (uf.zero) {
    assert(uf.exponent == 0x0);
    assert(uf.significand == 0x0);

  } else if (uf.subnormal) {
    assert(uf.exponent < -126);
    assert(uf.significand != 0x0);

    int biased = (uf.exponent + BIAS) - 1;

    assert(biased <= 0);
    assert(biased >= -23);
    assert((uf.significand & 0x800000) == 0x800000);  // Normalised
    assert((uf.significand & ((1 << -biased) - 1)) == 0x0); // Correctly rounded

  } else {
    assert(uf.exponent >= -126);
    assert(uf.exponent <=  127);
    assert(uf.significand != 0x0);
    assert((uf.significand & 0x800000) == 0x800000);

  }

  return;
}


// Normalises a denormal number or deal with catastrophic cancelation
void normaliseUp (unpackedFloat *uf) {
  int shift = 0;

  if ((0xFFFF00 & uf->significand) == 0) {
    uf->significand <<= 16;
    shift += 16;
  }
  if ((0xFF0000 & uf->significand) == 0) {
    uf->significand <<= 8;
    shift += 8;
  }
  if ((0xF00000 & uf->significand) == 0) {
    uf->significand <<= 4;
    shift += 4;
  }
  if ((0xC00000 & uf->significand) == 0) {
    uf->significand <<= 2;
    shift += 2;
  }
  if ((0x800000 & uf->significand) == 0) {
    uf->significand <<= 1;
    shift += 1;
  }

  uf->exponent -= shift;
}


unpackedFloat unpack_float (float f) {
  unpackedFloat uf;
  mix m;

  m.f = f;

  initUnpackedFloat(&uf);
  uf.sign = m.b.sign;

  if (m.b.exponent == 0) {
    if (m.b.significand == 0) {
      makeZero(&uf);

    } else {
      uf.nan = 0;
      uf.inf = 0;
      uf.zero = 0;
      uf.subnormal = 1;

      uf.exponent = -126;
      uf.significand = m.b.significand;

      normaliseUp(&uf);
    }

  } else if (m.b.exponent == 0xFF) {
    if (m.b.significand == 0) {
      makeInf(&uf);
    } else {
      makeNaN(&uf);
    }

  } else {
    // Normal

    uf.nan = 0;
    uf.inf = 0;
    uf.zero = 0;
    uf.subnormal = 0;

    uf.exponent = m.b.exponent - BIAS;
    uf.significand = 0x800000 | m.b.significand;
  }

  return uf;
}
